/* ----------------------------------------------------------------------
   The SINUS Firewall -- a TCP/IP packet filter for Linux
   
   Written within the SINUS project at the University of Zurich,
   SWITCH, Telekurs Payserv AG, ETH Zurich.
   originally based on the sf Firewall Software (C) 1996 by Robert
   Muchsel and Roland Schmid.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   SINUS Firewall resources:
   SINUS Homepage: http://www.ifi.unizh.ch/ikm/SINUS/
   Firewall Homepage: http://www.ifi.unizh.ch/ikm/SINUS/firewall.html
   Frequently asked questions: http://www.ifi.unizh.ch/ikm/SINUS/sf_faq.html
   Mailing list for comments, questions, bug reports: firewall@ifi.unizh.ch
   ----------------------------------------------------------------------  */

/*
 * In this file all data structures are defined that are used
 * by the firewall daemon and by the filter function.
 *
 * $Id: sf_global.h,v 1.19 1995/09/12 15:08:59 robby Rel $
 *
 */

#ifndef _SF_GLOBAL_H
#define _SF_GLOBAL_H

#define SF_VERSION "0.1.4"


#ifdef __KERNEL__

#include <linux/ip.h> 

#else

/* Due to incompatibilities between kernel headers and libc header,
 * we need some definitions here from the kernel headers. These ones
 * are taken from Linux-2.2.5. Be aware that they may change with a
 * new kernel version.  -hw
 */

#include <netinet/ip.h>
#define __u8 unsigned char
#define __u16 unsigned short
#define __u32 unsigned int

/* from <linux/timer.h> */
struct timer_list {
        struct timer_list *next;
        struct timer_list *prev;
        unsigned long expires;
        unsigned long data;
        void (*function)(unsigned long);
};

/* from <linux/firewall.h> */
#define FW_QUEUE        0
#define FW_BLOCK        1
#define FW_ACCEPT       2
#define FW_REJECT       (-1)
#define FW_REDIRECT     3
#define FW_MASQUERADE   4
#define FW_SKIP         5


#endif

#ifndef MAX
#define MAX(a, b)   ((a) > (b) ? (a) : (b))
#endif

/*
 *      Format of an IP firewall descriptor
 */

struct sf_fw {
#define SF_FW_MAGIC 0x524F524FL	/* RORO */
  unsigned long magic;		/* Magic value */

  struct sf_fw *fw_next;	/* Next firewall on chain */

  int fw_line;			/* line number of rule */
  __u32 fw_ipsrc;		/* source host of rule */

  union {
    long num;			/* identification of entry */
    struct sf_fw *ptr;		/* filter uses num, daemon uses ptr */
  } rule_id;

  union {			/* union to save storage */
    struct notification *ptr;	/* pointer to notification level */
    int num;			/* notification level for rule */
  } level;			/* num is only valid during init */

#define SF_RULE_CONNECTION		-1
#define SF_RULE_STATIC                  -2
#define SF_RULE_SPOOFING                -3
#define SF_RULE_OVERSIZED		-4
#define SF_RULE_REJECT_FTP		-5
  long timeout;			/* state of rule (active, ...) */

  int fw_rc;			/* return code on match (see sf_rc.h) */

  int fw_src_idx, fw_dst_idx,	/* Index to address array for */
      fw_rip_idx, fw_src_cnt,	/* source and dest. IP addresses, */
      fw_dst_cnt, fw_rip_cnt;	/* index to rip addresses */
  int fw_dyn_src:1, 
      fw_dyn_dst:1,
      fw_dyn_rip:1;

  __u8 ttl;			/* time to live value to compare with */
  __u8 protocol;		/* protocol */

  unsigned long fw_flags;	/* see below */

  unsigned long usage;		/* usage counter */
  unsigned long bytes;		/* byte usage counter */
  unsigned long gbytes;		/* gigabyte usage counter */
};

#define GIGA 1024*1024*1024

/*
 *    Values for "flags" field:

 *       bit         meaning
 *        0           log packet (notification_level > 0)
 *        1           check ttl field
 *        2-3         ttl: 00 -> ttl == value
 *                         01 -> ttl < value
 *                         10 -> ttl > value
 *                         11 -> ttl != value
 *        4           1 == match "from" if none of the specified 
 *                         addresses match (used for "outside")
 *        5           1 == match "to" if none of the specified
 *                         addresses match (used for "outside")
 *        6           option record_route
 *        7           option timestamp
 *        8           option security
 *        9           option loose_source_route
 *       10           option strict_source_route
 *       11           option sat_id
 *       12-15        protocol: 0000  all protocols
 *                              0001  rip
 *                              1111  current protocol (dynamic rule)
 *       16-31        type (protocol dependent) 
 */

/* structure of address array entry */

#define SF_ADDRESS_CNT_MAX   340	/* 340 * 12 bytes == 4080 bytes */
/* SF_ADDRESS_CNT_MAX can now savely set to higher values due to
 * the introduction of the longatomicread function in sf_sfc.c
 * since sifi-0.1.3. --hw */
struct sf_address {
  __u32 addr, mask;		/* ip address and mask (this is NOT the netmask) */
  /* (both in network byte order) */
  unsigned short port, prend;	/* port, end of port range (host byte order) */
  /* if prend is 0 and port == USHRT_MAX, the address entry is free */
};

#define SF_IGNORE_ADDRESS     0	/* match on all ip addresses */
#define SF_SOURCEHOST        -1	/* match source address in dynamic rule */
#define SF_SOURCENET         -2	/* match source net in dynamic rule */
#define SF_DESTHOST          -3	/* match destination address in dynamic rule */
#define SF_DESTNET           -4	/* match destination net in dynamic rule */

#define SF_FW_LOG            0x0001	/* log packet */

#define SF_FW_CHECK_TTL      0x0002	/* check time to live field */
#define SF_FW_TTL            0x000E	/* mask for SF_FW_TTL */
#define SF_FW_TTL_EQUAL      0x0002	/* ttl == value */
#define SF_FW_TTL_LESS       0x0006	/* ttl < value */
#define SF_FW_TTL_GREATER    0x000A	/* ttl > value */
#define SF_FW_TTL_NOTEQUAL   0x000E	/* ttl != value */

#define SF_FW_SRC_NEG	     0x0010	/* from condition negated */
#define SF_FW_DST_NEG	     0x0020	/* to condition negated */

#define SF_FW_CHECK_OPT	     0x0FC0	/* check for options */
#define SF_FW_OPT_RR         0x0040	/* record route */
#define SF_FW_OPT_TS         0x0080	/* time stamp */
#define SF_FW_OPT_SEC        0x0100	/* security */
#define SF_FW_OPT_LSR        0x0200	/* loose source route */
#define SF_FW_OPT_SSR        0x0400	/* strict source route */
#define SF_FW_OPT_SATID      0x0800	/* sat id */

#define SF_FW_PROT           0xF000	/* mask to isolate protocol */
#define SF_FW_PROT_ALL       0x0000	/* matches all protocols */
#define SF_FW_PROT_RIP       0x1000	/* rip */
#define SF_FW_CHECK_PROTOCOL 0xE000	/* compare protocol field */
#define SF_FW_PROT_CURRENT   0xF000	/* match current protocol in dynamic rule */

/* definitions for protocol dependent type bits: */
#define SF_TYPE_MASK           0xFFFF0000	/* mask to isolate protcol dependent bits */

#define SF_FTP_DATA_LOG        0x00010000	/* log data connections (TCP and ALL only) */
#define SF_FTP_NO_ACTIVE       0x00020000	/* don't allow FTP active mode  */
#define SF_FTP_NO_PASSIVE      0x00040000	/* don't allow FTP passive mode */

#define SF_CHECK_SEQUENCE      0x00080000	/* check TCP sequence numbers   */

#define SF_ICMP_ALLTYPES       0x00000000
#define SF_ICMP_ECHOREPLY      0x00010000	/* Echo Reply                   */
#define SF_ICMP_DEST_UNREACH   0x00020000	/* Destination Unreachable      */
#define SF_ICMP_SOURCE_QUENCH  0x00040000	/* Source Quench                */
#define SF_ICMP_REDIRECT       0x00080000	/* Redirect (change route)      */
#define SF_ICMP_ECHO           0x00100000	/* Echo Request                 */
#define SF_ICMP_TIME_EXCEEDED  0x00200000	/* Time Exceeded                */
#define SF_ICMP_PARAMETERPROB  0x00400000	/* Parameter Problem            */
#define SF_ICMP_TIMESTAMP      0x00800000	/* Timestamp Request            */
#define SF_ICMP_TIMESTAMPREPLY 0x01000000	/* Timestamp Reply              */
#define SF_ICMP_INFO_REQUEST   0x02000000	/* Information Request          */
#define SF_ICMP_INFO_REPLY     0x04000000	/* Information Reply            */
#define SF_ICMP_ADDRESS        0x08000000	/* Address Mask Request         */
#define SF_ICMP_ADDRESSREPLY   0x10000000	/* Address Mask Reply           */

#define SF_IGMP_ALLTYPES               0x00000000
#define SF_IGMP_HOST_MEMBERSHIP_QUERY  0x00010000	/* From RFC1112 */
#define SF_IGMP_HOST_MEMBERSHIP_REPORT 0x00020000	/* Ditto */
#define SF_IGMP_HOST_LEAVE_MESSAGE     0x00040000	/* An extra BSD seems to send */

#define SF_RIP_ADDR_NEG        0x00010000	/* match rip-addr if none of the specified */
					   /* addresses match (used for "outside") */

struct sf_fw_ip {
  struct sf_fw rule;
  struct iphdr ip;
  unsigned short checksum;
  unsigned char static_addr[0];
};

#define PROC_SIZE_IP            176

#define RULE_SPOOF_RECV         -1      /* packet received on wrong interface*/
#define RULE_SPOOF_XMIT         -2      /* packet sent via wrong interface */
#define RULE_OVERSIZED          -3      /* oversized packet received */
#define RULE_CONNECTION         -4      /* ftp data connection */
#define RULE_REJECT_FTP         -5      /* reject ftp data connection */

struct sf_proc_entry {
  int action;                   /* action code (= return code)  */
  union {
    long num;                   /* rule number                  */
    struct sf_fw *ptr;          /* pointer to rule              */
  } rule;
  char devname[8];              /* device address               */
  char ip_packet[PROC_SIZE_IP]; /* ip packet (first N bytes)    */
};

#define NUM_PROC_ENTRIES        21

struct sf_proc {
  unsigned num_entries;         /* number of valid records      */
  unsigned long lost_entries;   /* number of lost packets       */
  struct sf_proc_entry data[NUM_PROC_ENTRIES];
};

#define FTP_DATA_PORT           20
#define FTP_CONTROL_PORT        21

/* TCP circuits */

/* The client is the host that initiates the connection. */
#define SF_TCP_ACCEPT_SYN       0x00    /* accept client SYN   */

#define SF_TCP_CLIENT_SYN       0x01    /* client has sent SYN */
#define SF_TCP_SERVER_SYN       0x02    /* server has sent SYN ACK */
#define SF_TCP_SYN_ACK          0x03    /* both of the above */

#define SF_TCP_CLIENT_ACK       0x04    /* client ACK of SYN ACK */
#define SF_TCP_ESTABLISHED3     0x07    /* connection is up */

#define SF_TCP_FTP              0x08    /* is ftp control session */
#define SF_TCP_ESTABLISHEDFTP   0x0F    /* ftp control connection is up */

#define SF_TCP_CLIENT_FIN       0x10    /* client has sent FIN */
#define SF_TCP_SERVER_FIN       0x20    /* server has sent FIN */
#define SF_TCP_TERMINATED       0x37    /* both of the above */

struct sf_tcp_connection {
  __u32 id;
  struct sf_tcp_connection *prev, *next;
  struct sf_tcp_connection *before, *after;
  __u32 client_addr, server_addr;
  __u16 client_port, server_port;
  __u16 client_win, server_win; /* window sizes */
  __u8 client_scale, server_scale; /* window scale factors */
  __u32 client_seq, server_seq; /* sequence numbers */
  int state;                    /* state of tcp connection */
  struct timer_list timer;
  __u32 established;            /* "jiffies" connection was established */
  __u32 lastused;               /* "jiffies" connection was last used */
  struct sf_fw *rule;           /* accepted/created by rule */
  struct device *rif;           /* device of initial SYN */
  int timer_set:1,              /* one if a timer is set */
    allpsource:1,             /* one if all source ports match */
    log_open:1,               /* log beginning of connection */
    log_dxfer:1,              /* log data transfers (i.e. FTP) */
    ftp_allow_active:1,       /* active mode FTP is allowed */
    ftp_allow_passive:1,      /* passive mode FTP is allowed */
    check_sequence:1;         /* check TCP sequence numbers */
};

#define SF_TCP_KEY(a)  ( (a) % SF_TCP_HASH_SIZE )
#define SF_TCP_HASH_SIZE        499     /* must be a prime number */

#endif				/* _SF_GLOBAL_H */
