/* ----------------------------------------------------------------------
   The SINUS Firewall -- a TCP/IP packet filter for Linux

   Written within the SINUS project at the University of Zurich,
   SWITCH, Telekurs Payserv AG, ETH Zurich.
   originally based on the sf Firewall Software (C) 1996 by Robert
   Muchsel and Roland Schmid.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   SINUS Firewall resources:
   SINUS Homepage: http://www.ifi.unizh.ch/ikm/SINUS/
   Firewall Homepage: http://www.ifi.unizh.ch/ikm/SINUS/firewall.html
   Frequently asked questions: http://www.ifi.unizh.ch/ikm/SINUS/sf_faq.html
   Mailing list for comments, questions, bug reports: firewall@ifi.unizh.ch
   ----------------------------------------------------------------------  */


/*
 * sf Firewall Daemon Startup/sfc User Control Module
 *
 * $Id: sfc.c,v 1.15 1995/09/12 15:08:59 robby Rel $
 */

#include <netdb.h>
#include <pwd.h>
#include <grp.h>
#if __GLIBC__ == 2
#include <signal.h>
#else
#include <bsd/signal.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <arpa/inet.h>
#include <asm/param.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <syscall.h>
#include <sys/sysinfo.h>
#include "sf_custom.h"
#include "sf_global.h"
#include "sf_config.h"
#include "sf_daemon.h"
#include "sf_kernel.h"
#include "sf_rc.h"
#include "sf_proto.h"
#include "sf_control.h"
#include "sf_server.h"
#include "sf_comm.h"

/* This one is missing from the headers!? */
/* _syscall1(int, sysinfo, struct sysinfo *, info); */

static int running(void);
static void showcfg_or_kill(int, __u32);
static int tcpkill(__u32, __u32, __u32);
static void tcpkillarg(char *, char *, char *);
static void tcpkillidle(char *);
static void tcpkillall(void);
static void resetusage(void);
static char *arg0;
static void print_addr(unsigned long, unsigned long, int, char *, int);
static char *print_state(int);
static char *print_time(time_t, time_t, time_t);
static char *print_diff(time_t, time_t);
static void start_loop(char *, int, struct passwd *);
int longatomicread (int, void *, int);

int sf_uid, sf_gid;

/*
 * Show the syntax.
 */

static void show_syntax(void)
{
  fprintf(stderr, "\nUsage: %s command, where command is one of:\n\n" \
          "start [config]         Start the firewall daemon and configure it.\n" \
	  "                       Use `config' to specify a configuration file\n" \
	  "                       (default: " DEFAULT_CONFIG ")\n\n" \
          "stop [force]           Signal the firewall daemon until it stops.\n" \
	  "                       Note! To disable the firewall, you must unload\n" \
	  "                       the kernel filter module.\n" \
          "                       `force' stops the daemon even if a remote configuration\n" \
          "                       client is connected to the daemon.\n\n" \
	  "reconfig [config]      Stop the firewall daemon, and restart it with a new\n" \
	  " [flush or flush_all]  config file (default: " DEFAULT_CONFIG ").\n" \
	  "                       `reconfig' does NOT stop any existing TCP connections,\n" \
          "                       unless `flush' or `flush_all' is specified.\n" \
	  "                       `flush' ends all existing TCP connections which are\n" \
          "                       not allowed under the new configuration.\n" \
	  "                       `flush_all' ends all existing TCP connections.\n\n" \
	  "checkconfig [config]   Check a configuration file for syntax errors.\n" \
	  "                       (default: " DEFAULT_CONFIG ")\n\n" \
	  "show                   Show current configuration and TCP connections.\n\n" \
	  "showtcp                Show TCP connections only.\n\n" \
	  "kill id key created    Kill specified TCP connection.\n\n" \
	  "killall                Kill all TCP connections.\n\n" \
          "killidle minutes       Kill all TCP connections idle for at least minutes.\n\n" \
	  "resetusage             Reset usage counters.\n\n", arg0);

  exit(1);
}


/*
 * killproc() - kill a running process 
 */

static void killproc(int force)
{
  int pid;
  int sig;

  if (!(pid = running())) {
    fprintf(stderr, "The firewall daemon is not running!\n");
    exit(1);
  }

  if (force)
    sig = SIGTERM;
  else
    sig = SIGUSR1;

  kill(pid, sig);
  sleep(1);
  if (running()) {
    fputs("The daemon is still running, waiting 5 seconds more...\n", stderr);
    sleep(3);
    kill(pid, sig);
    sleep(2);
    if (running()) {
      if (force) {
        fputs("The daemon is still running, sending SIGKILL.\n", stderr);
        kill(pid, SIGKILL);
      }
      else {
        fputs("The daemon is still running, probably because a remote configuration\n"
              "connection is established. Use `sfc stop force' if you want to stop\n"
              "the daemon anyway. Reconfiguration is not allowed.\n", stderr);
        exit(1);
      }
    }
  }
}


/*
 * main() - check command line arguments and process them 
 */

#define A_STOP      0x01
#define A_START     0x02
#define A_FLUSH     0x04
#define A_FLUSH_ALL 0x08
#define A_SHOW      0x10
#define A_SHOWTCP   0x20
#define A_CHECKCFG  0x40
#define A_TCP_KILL  0x80
#define A_TCP_KILLA 0x100
#define A_TCP_KILLI 0x200
#define A_FORCE     0x400
#define A_RESETUSE  0x800

int main(int argc, char **argv)
{
  char *config = DEFAULT_CONFIG;
  int action;
  struct passwd *pwd = NULL;
  struct group *grp = NULL;
  struct stat iostat;
  FILE *spyfd;

  /* check sanity first */
  if (sizeof(struct sf_tcp_connection) == sizeof(struct sf_fw)) {
    fputs("Internal error - struct sizes equal, change sf_fw or sf_tcp_conn!\n", stderr);
    exit(1);
  }

  /* check if unprivleged user and group exists */
  if (((sf_uid = getuid()) == 0) || ((sf_uid = geteuid()) == 0)) {
    if (!(pwd = getpwnam(UNPRIV_NAME)) || (pwd->pw_uid <= 0)) {
      perror("Please create an unprivileged user named `firewall'");
      exit(1);
    }
    sf_uid = pwd->pw_uid;
  }
  if (((sf_gid = getgid()) == 0) || ((sf_gid = getegid()) == 0)) {
    if (!(grp = getgrnam(UNPRIV_NAME)) || (grp->gr_gid <= 0)) {
      perror("Please create an unprivileged group named `firewall'");
      exit(1);
    }
    sf_gid = grp->gr_gid;
  }

  arg0 = argv[0];

  /* check arguments */
  if (argc < 2)
    show_syntax();

  if (strcmp("start", argv[1]) == 0) {
    if (argc > 3)
      show_syntax();
    else if (argc == 3)
      config = argv[2];
    action = A_START;
  }
  else if (strcmp("stop", argv[1]) == 0) {
    if (argc > 3)
      show_syntax();
    action = A_STOP;
    if (argc == 3) {
      if (strcmp("force", argv[2]) == 0)
        action |= A_FORCE;
      else show_syntax();
    }
  }
  else if ((strcmp("reconfig", argv[1]) == 0)
	   || (strcmp("restart", argv[1]) == 0)) {
    action = A_STOP | A_START | A_RESETUSE;
    if (argc > 4)
      show_syntax();
    else if (argc == 3) {
      if (strcmp(argv[2], "flush") == 0)
	action |= A_FLUSH;
      else if (strcmp(argv[2], "flush_all") == 0)
	action |= A_FLUSH_ALL;
      else
	config = argv[2];
    }
    else if (argc == 4) {
      if (strcmp(argv[2], "flush") == 0) {
	action |= A_FLUSH;
	config = argv[3];
      }
      else if (strcmp(argv[2], "flush_all") == 0) {
	action |= A_FLUSH_ALL;
	config = argv[3];
      }
      else if (strcmp(argv[3], "flush") == 0) {
	action |= A_FLUSH;
	config = argv[2];
      }
      else if (strcmp(argv[3], "flush_all") == 0) {
	action |= A_FLUSH_ALL;
	config = argv[2];
      }
      else
	show_syntax();
    }
  }
  else if (strcmp("show", argv[1]) == 0) {
    if (argc != 2)
      show_syntax();
    action = A_SHOW;
  }
  else if (strcmp("showtcp", argv[1]) == 0) {
    if (argc != 2)
      show_syntax();
    action = A_SHOWTCP;
  }
  else if (strcmp("kill", argv[1]) == 0) {
    if (argc != 5)
      show_syntax();
    action = A_TCP_KILL;
  }
  else if (strcmp("killall", argv[1]) == 0) {
    if (argc != 2)
      show_syntax();
    action = A_TCP_KILLA;
  }
  else if (strcmp("killidle", argv[1]) == 0) {
    if (argc != 3)
      show_syntax();
    action = A_TCP_KILLI;
  }
  else if (strcmp("resetusage", argv[1]) == 0) {
    if (argc != 2)
      show_syntax();
    action = A_RESETUSE;
  }
  else if ((strcmp("checkconfig", argv[1]) == 0)
	   || (strcmp("check", argv[1]) == 0)) {
    if (argc > 3)
      show_syntax();
    else if (argc == 3)
      config = argv[2];
    action = A_CHECKCFG;
  }
  else
    show_syntax();

  /* check for pipes and create them if neccessary --hw */
  if(stat(IPIPE, &iostat) != 0                              /* no input pipe */
     || !S_ISFIFO(iostat.st_mode)                     /* exists, but no pipe */
     || iostat.st_uid != sf_uid                           /* wrong ownership */
     || iostat.st_gid != sf_gid
     || (iostat.st_mode | S_IRUSR) == 0                    /* no read access */
     || (iostat.st_mode | S_IWUSR) == 0                   /* no write access */
     || (iostat.st_mode & (S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH)) != 0)
                                                     /* too much permissions */
    {
      mknod(IPIPE, S_IFIFO|S_IRUSR|S_IWUSR, 0);
      chown(IPIPE, sf_uid, sf_gid);
    }
  if((stat(OPIPE, &iostat) != 0 && errno == ENOENT)        /* no output pipe */
     || !S_ISFIFO(iostat.st_mode)                     /* exists, but no pipe */
     || iostat.st_uid != sf_uid                           /* wrong ownership */
     || iostat.st_gid != sf_gid
     || (iostat.st_mode | S_IRUSR) == 0                    /* no read access */
     || (iostat.st_mode | S_IWUSR) == 0                   /* no write access */
     || (iostat.st_mode & (S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH)) != 0)
                                                     /* too much permissions */
    {
      mknod(OPIPE, S_IFIFO|S_IRUSR|S_IWUSR, 0);
      chown(OPIPE, sf_uid, sf_gid);
    }

  if(stat(SPY_TMP, &iostat) != 0                        /* no temporary file */
     || !S_ISREG(iostat.st_mode)                       /* not a regular file */
     || iostat.st_uid != sf_uid                           /* wrong ownership */
     || iostat.st_gid != sf_gid
     || (iostat.st_mode | S_IRUSR) == 0                    /* no read access */
     || (iostat.st_mode | S_IWUSR) == 0                   /* no write access */
     || (iostat.st_mode & (S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH)) != 0)
                                                     /* too much permissions */
    {
      spyfd = fopen(SPY_TMP, "w");
      if(spyfd == NULL) {
	fprintf(stderr, "Error opening the temp file `%s'!\n", SPY_TMP);
	exit(1);
      }
      fclose(spyfd);
      chmod(SPY_TMP, S_IRUSR | S_IWUSR);
      chown(SPY_TMP, sf_uid, sf_gid);
    }
  
  if ((action & A_RESETUSE) == A_RESETUSE)
    resetusage();

  /* create a new process? */
  if ((action & A_START) == A_START)
    start_loop(config, action, pwd); /* does not return */

  /* parse config file first */
  if ((action & A_CHECKCFG) == A_CHECKCFG) {
    if (sf_read_config_file(config) != 0) {
      fprintf(stderr, "Error parsing the configuration file `%s'!\n", config);
      exit(1);
    }
    else 
      fprintf(stderr, "Parsing `%s': OK\n", config);
  }


  /* kill the process if requested */
  if ((action & A_STOP) == A_STOP) {
    killproc((action & A_FORCE) == A_FORCE);
    unlink(PID_FILE);
  }


  if ((action & A_SHOW) == A_SHOW)
    showcfg_or_kill(0, 0);

  if ((action & A_SHOWTCP) == A_SHOWTCP)
    showcfg_or_kill(1, 0);

  if ((action & A_TCP_KILL) == A_TCP_KILL)
    tcpkillarg(argv[2], argv[3], argv[4]);

  if ((action & A_TCP_KILLA) == A_TCP_KILLA)
    tcpkillall();

  if ((action & A_TCP_KILLI) == A_TCP_KILLI)
    tcpkillidle(argv[2]);

  return 0;
}


/*
 *
 */

void start_loop(char *config, int action, struct passwd *pwd)
{
  int i, pid, fd, log, status;
  char buf[20];

  printf("Initializing daemon...\n");

  strcpy(current_config, config);

  if ((action & A_STOP) == A_STOP) {
    if ((pid = fork()) == 0) {
      memset(sf_addr, 0, sizeof(sf_addr));
      for (i = 0; i < SF_ADDRESS_CNT_MAX; i++)
        sf_addr[i].port = USHRT_MAX;
      exit(sf_read_config_file(config));
    }
    if (pid < 0) {
      perror("Cannot fork()");
      exit(1);
    }
    waitpid(pid, &status, 0);

    /* check exit code */
    if (!WIFEXITED(status) || (WEXITSTATUS(status) != 0)) {
      fprintf(stderr, "Error parsing the configuration file `%s'!\n", config);
      exit(WEXITSTATUS(status));
    }

    killproc((action & A_FORCE) == A_FORCE);
  }

  /* check whether process is already or still running */
  if ((pid = running()) != 0) {
    fprintf(stderr, "The firewall daemon is already running (pid %i)!\n", pid);
    exit(1);
  }
  unlink(PID_FILE);

  for (i = 1; i < NSIG; i++)
    signal(i, SIG_IGN);

  if ((pid = fork()) == 0) {	/* child */
    /* become a daemon */
    setsid();     /* disconnect from the terminal */
    chdir("/");   /* just in case someone wants to unmount the current dir. */

    while (1) {
      /* create the /var/run/firewall.pid file and lock it */
      if ((fd = creat(PID_FILE, 0600)) < 0) {
        perror("Cannot create pid file");
        exit(1);
      }
      if ((getuid() == 0) || (geteuid() == 0))
        fchown(fd, pwd->pw_uid, pwd->pw_gid);
      if ((flock(fd, LOCK_EX | LOCK_NB)) < 0) {
        perror("Cannot create lock");
        exit(1);
      }
      
      if ((pid = fork()) == 0) {	/* child */
        close(fd);

        memset(sf_addr, 0, sizeof(sf_addr));
        for (i = 0; i < SF_ADDRESS_CNT_MAX; i++)
          sf_addr[i].port = USHRT_MAX;

        /* parse config file */
        if (sf_read_config_file(config) != 0) {
          fprintf(stderr, "Error parsing the configuration file `%s'!\n", config);
          exit(1);
        }

        /* create log files and change owner if possible or if needed */
        if ((log = open(LOG, O_CREAT | O_APPEND | O_RDWR, 0600)) < 0) {
          perror("Cannot open log file");
          exit(1);
        }
        if ((getuid() == 0) || (geteuid() == 0))
          fchown(log, pwd->pw_uid, pwd->pw_gid);
        close(log);

        if ((log = open(SPY_LOG, O_CREAT | O_APPEND | O_RDWR, 0600)) < 0) {
          perror("Cannot open spy log file");
          exit(1);
        }
        if ((getuid() == 0) || (geteuid() == 0))
          fchown(log, pwd->pw_uid, pwd->pw_gid);
        close(log);

        if ((log = open(REPORT, O_CREAT | O_APPEND | O_RDWR, 0600)) < 0) {
          perror("Cannot open report file");
          exit(1);
        }
        if ((getuid() == 0) || (geteuid() == 0))
          fchown(log, pwd->pw_uid, pwd->pw_gid);
        close(log);

        init_log();

        /* disconnect tcp sessions which are not allowed under the new confiuration */
        if ((action & A_FLUSH) == A_FLUSH)
          sf_config_flush(fddev);

        /* disconnect all tcp sessions */
        if ((action & A_FLUSH_ALL) == A_FLUSH_ALL)
          sf_config_flush_all(fddev);

        fdlock = open(PID_FILE, O_RDWR);
        flock(fdlock, LOCK_EX);	/* wait until parent has given up the lock */

        /* give up super user privileges */
        if ((getuid() == 0) || (geteuid() == 0)) {
	  setgid(pwd->pw_gid);
	  setuid(pwd->pw_uid);	/* how sad - all priviliges lost */
	}

        /* start the daemon - this call should not return */
        start_log(config);

        exit(2);			/* but you never know */
      }

      if (pid < 0) {
        perror("Cannot fork()");
        exit(1);
      }

      /* record the child's pid */
      sprintf(buf, "%i\n", pid);
      if (write(fd, buf, strlen(buf)) < 0) {
        perror("Cannot write to lock file");
        exit(1);
      }

      /* give up lock so child will run */
      flock(fd, LOCK_UN);
      close(fd);

      /* wait for child exit */
      waitpid(pid, &status, 0);

      /* restart child on unhandled signal (SIGFAULT, ...) 
         or exit() with exit code != 0 */
      if (!WIFSIGNALED(status) && WIFEXITED(status) && 
          ((WEXITSTATUS(status) == 0) || (WEXITSTATUS(status) == 1)))
        exit(WEXITSTATUS(status)); /* clean exit */

      sleep(5); /* avoid trashing if continuous crashing */
      action = A_START;
    } /* while */

    exit(1);
  }

  if (pid < 0) {
    perror("Cannot fork()");
    exit(1);
  }

  exit(0);
}


/*
 * See if daemon is running and fetch the pid. This function uses
 * a pid file, which is locked by the daemon if it is alive.
 */

int running(void)
{
  struct stat statbuf;
  int fd, pid;
  FILE *fdf;

  /* does the pid file exist? */
  if (stat(PID_FILE, &statbuf) < 0)
    return 0;			/* no, the daemon can't be running */

  /* is it a regular file? */
  if (!S_ISREG(statbuf.st_mode)) {
    fprintf(stderr, "The file %s is not a regular file. sfc cannot continue.\n", PID_FILE);
    exit(1);
  }

  /* open the file and lock it */
  if ((fd = open(PID_FILE, O_RDWR)) < 0) {
    fprintf(stderr, "Cannot open the file %s. sfc cannot continue.\n", PID_FILE);
    exit(1);
  }

  /* get lock */
  if (flock(fd, LOCK_NB | LOCK_EX) >= 0) {	/* ok - lock process cannot be running */
    flock(fd, LOCK_NB | LOCK_UN);
    close(fd);
    return 0;
  }
  close(fd);

  /* didn't get lock, so the daemon is alive (cross your fingers) */

  /* read pid from file */
  if ((fdf = fopen(PID_FILE, "r")) == NULL) {
    fprintf(stderr, "Cannot open the file %s. sfc cannot continue.\n", PID_FILE);
    exit(1);
  }

  pid = 0;
  fscanf(fdf, "%i", &pid);
  fclose(fdf);

  return pid;
}


/*
 * Data structures and functions for pointer conversion: daemon to sfc
 */

#define M 277			/* some prime number */
#define hash(p) (((unsigned) (p)) % M)

struct hashlist {
  void *old;
  void *new;
  struct hashlist *next;
};

struct hashlist ahash[M];

void *convertp(void *p)
{				/* converts a pointer from old to new */
  struct hashlist *hl;

  if (p == NULL)
    return NULL;

  hl = &(ahash[hash(p)]);

  while ((hl != NULL) && (hl->old != p))
    hl = hl->next;

  if (hl == NULL) {
    fprintf(stderr, "Upps... Cannot convert this pointer!?!\n");
    exit(1);
  }

  return hl->new;
}

void addp(void *old, void *new)
{				/* add pointer to the hash structure */
  unsigned int h;

  if (old == NULL)
    return;

  h = hash(old);
  if (ahash[h].old == NULL) {
    ahash[h].old = old;
    ahash[h].new = new;
  }
  else {
    struct hashlist *hl;

    hl = &(ahash[h]);
    while (hl->next != NULL)
      hl = hl->next;
    hl->next = malloc(sizeof(struct hashlist));

    hl = hl->next;
    hl->old = old;
    hl->new = new;
    hl->next = NULL;
  }
}

/*
 * Show the current configuration 
 */

static void showcfg_or_kill(int tcponly, __u32 idlekill)
{
  char buffer[255];
  struct sf_fw *rules, *r;
  int fd, result, nrule, i;
  time_t ltime, now, usage_since = 0;
  struct sysinfo info;
  int pid = 0;

  memset(ahash, 0, sizeof(ahash));

  if (!tcponly) {
    if ((pid = running()) != 0) {

      /* First get the data from the daemon */
      kill(pid, SIGUSR2);	/* signal the process to fetch/write data */

      if ((fd = open(IPIPE, O_WRONLY)) < 0) {
	perror("Cannot open the firewall input pipe");
	exit(1);
      }
      i = SF_CMD_DAEMON_MAGIC;
      write(fd, &i, sizeof(i));
      i = SF_CMD_DAEMON_RULES;
      write(fd, &i, sizeof(i));
      close(fd);

      if ((fd = open(OPIPE, O_RDONLY)) < 0) {
	perror("Cannot open the firewall output pipe");
	exit(1);
      }

      /* get usage since */
      if (longatomicread(fd, &usage_since, sizeof(usage_since)) < 0)
        perror("read");

      /* get the address array */
      if (longatomicread(fd, &sf_addr_free, sizeof(sf_addr_free)) < 0)
        perror("read");
      if (longatomicread(fd, sf_addr, sizeof(sf_addr)) < 0) 
        perror("read");

      rules = NULL;

      /* get all rules */
      if (longatomicread(fd, &r, sizeof(r)) < 0)
        perror("read");
      while (r != NULL) {
	if (rules == NULL)
	  rules = malloc(sizeof(struct sf_fw));
	else {
	  rules->fw_next = malloc(sizeof(struct sf_fw));
	  rules = rules->fw_next;
	}
	addp(r, rules);

	if (longatomicread(fd, rules, sizeof(*rules)) < 0)
          perror("read");
	if (longatomicread(fd, &r, sizeof(r)) < 0)
          perror("read");
      }
      if (rules != NULL)
	rules->fw_next = NULL;

      /* read variables and values */
      if (longatomicread(fd, &sf_var_free, sizeof(sf_var_free)) < 0)
        perror("read");
      if (longatomicread(fd, variables, sizeof(variables)) < 0)
        perror("read");

      for (i = 0; i < SF_VAR_CNT_MAX; i++) {
	struct variable *v, *w;

	w = &(variables[i]);

	longatomicread(fd, &v, sizeof(v));
	while (v != NULL) {
	  w->next = malloc(sizeof(struct variable));

	  w = w->next;

	  if (longatomicread(fd, w, sizeof(*w)) < 0)
            perror("read");
	  if (longatomicread(fd, &v, sizeof(v)) < 0)
            perror("read");
	}
	w->next = NULL;
      }

      for (i = 0; i < SF_VAR_CNT_MAX; i++) {
	char buffer[4096];
	char c, *bufp;

	bufp = buffer;
	do {
	  longatomicread(fd, &c, sizeof(c));
	  *bufp = c;
	  bufp++;
	} while (c != 0);
	if (*buffer)
	  varnames[i] = strdup(buffer);
	else
	  varnames[i] = NULL;
      }

      close(fd);
    }				/* daemon running */
    else {
      printf("Note: The daemon is not running - showing TCP data only.\n");
    }
  }


  /* Now get data from the kernel module */

  if ((fd = open(sf_devname, O_RDONLY)) < 0) {
    if (errno == ENODEV)
      fputs("The module sf.o is not loaded!\n", stderr);
    else
      perror("Cannot open the device for reading");
    exit(1);
  }

  if (pid && !tcponly) {	/* daemon running */
    printf("Active rules");
    if (usage_since)
      printf(" -- usage counters last reset at %s", ctime(&usage_since));
    else
      printf("\n");
    printf("------------\n");

    nrule = 0;
    do {
      struct sf_fw krule;

      result = longatomicread(fd, &krule, sizeof(krule));
      if (result < 0) {
	perror("Error reading the device");
	exit(1);
      }
      if (result > 0) {
	struct sf_fw *rule;
	int tcp_or_udp = 0;

	nrule++;
	rule = convertp(krule.rule_id.ptr);

	printf("\n%3i (%s%s%i) ", nrule, 
               krule.fw_ipsrc ? 
                 inet_ntoa(*(struct in_addr *) &(krule.fw_ipsrc)) : "", 
               krule.fw_ipsrc ? " l": "line ", 
               krule.fw_line);

	if (krule.timeout == SF_RULE_STATIC)
	  printf("static");
	else {
	  strftime(buffer, sizeof(buffer), TIME_FORMAT, localtime(&(rule->timeout)));
	  printf("dynamic, timeout %s", buffer);
	}
	printf(" -- usage: %lu time%s", krule.usage, krule.usage != 1 ? "s" : "");

        if (krule.gbytes > 0)
          printf(", %lu GB %lu b", krule.gbytes, krule.bytes);
        else
          if (krule.bytes > 0)
            printf(", %lu bytes", krule.bytes);

	printf("\n    ");
	switch (krule.fw_rc) {
	  case FW_ACCEPT:
	    printf("accept");
	    break;
          case SF_RC_OBSERVE:
            printf("observe");
            break;
	  case SF_RC_TREJECT:
	    printf("reject (tcp reset/best)");
	    break;
	  case SF_RC_ECHO:
	    printf("reject (echo reply)");
	    break;
	  case FW_BLOCK:
	    printf("block");
	    break;
	  case SF_RC_RNET:
	    printf("reject (net unreachable)");
	    break;
	  case FW_REJECT:
	    printf("reject (host unreachable)");
	    break;
	  case SF_RC_RPROTO:
	    printf("reject (protocol unreachable)");
	    break;
	  case SF_RC_RPORT:
	    printf("reject (port unreachable)");
	    break;
	}
	printf(", notification level %i\n", ((krule.fw_flags & SF_FW_LOG) == 0) ? 0 : rule->level.num);

	if ((krule.fw_flags & SF_FW_PROT) != SF_FW_PROT_ALL) {
	  if ((krule.fw_flags & SF_FW_PROT) == SF_FW_PROT_RIP) {
	    printf("    protocol RIP");
	    if (krule.fw_rip_cnt > 0) {
	      printf(", ");
	      if ((krule.fw_flags & SF_RIP_ADDR_NEG) != 0) {
		printf("not ");
		print_addr(krule.fw_rip_idx, krule.fw_rip_cnt,
			   0, "                   not ", 0);
	      }
	      else
		print_addr(krule.fw_rip_idx, krule.fw_rip_cnt,
			   0, "                   ", 0);
	    }
	    else
	      printf("\n");
	  }
	  else {
	    if (krule.protocol == IPPROTO_ICMP) {
	      printf("    protocol ICMP");
	      if ((krule.fw_flags & SF_TYPE_MASK) != SF_ICMP_ALLTYPES) {
		if ((krule.fw_flags & SF_ICMP_ECHOREPLY) != 0)
		  printf(" `echo reply'");
		if ((krule.fw_flags & SF_ICMP_DEST_UNREACH) != 0)
		  printf(" `dest. unreachable'");
		if ((krule.fw_flags & SF_ICMP_SOURCE_QUENCH) != 0)
		  printf(" `source quench'");
		if ((krule.fw_flags & SF_ICMP_REDIRECT) != 0)
		  printf(" `redirect'");
		if ((krule.fw_flags & SF_ICMP_ECHO) != 0)
		  printf(" `echo request'");
		if ((krule.fw_flags & SF_ICMP_TIME_EXCEEDED) != 0)
		  printf(" `time exceeded'");
		if ((krule.fw_flags & SF_ICMP_PARAMETERPROB) != 0)
		  printf(" `parameter problem'");
		if ((krule.fw_flags & SF_ICMP_TIMESTAMP) != 0)
		  printf(" `timestamp request'");
		if ((krule.fw_flags & SF_ICMP_TIMESTAMPREPLY) != 0)
		  printf(" `timestamp reply'");
		if ((krule.fw_flags & SF_ICMP_INFO_REQUEST) != 0)
		  printf(" `information request'");
		if ((krule.fw_flags & SF_ICMP_INFO_REPLY) != 0)
		  printf(" `information reply'");
		if ((krule.fw_flags & SF_ICMP_ADDRESS) != 0)
		  printf(" `address mask request'");
		if ((krule.fw_flags & SF_ICMP_ADDRESSREPLY) != 0)
		  printf(" `address mask reply'");
	      }
	      printf("\n");
	    }
	    else if (krule.protocol == IPPROTO_IGMP) {
	      printf("    protocol IGMP");
	      if ((krule.fw_flags & SF_TYPE_MASK) != SF_IGMP_ALLTYPES) {
		if ((krule.fw_flags & SF_IGMP_HOST_MEMBERSHIP_QUERY) != 0)
		  printf(" `membership query'");
		if ((krule.fw_flags & SF_IGMP_HOST_MEMBERSHIP_REPORT) != 0)
		  printf(" `membership report'");
		if ((krule.fw_flags & SF_IGMP_HOST_LEAVE_MESSAGE) != 0)
		  printf(" `leave message'");
	      }
	      printf("\n");
	    }
	    else if (krule.protocol == IPPROTO_TCP) {
              printf("    protocol %i (TCP)", krule.protocol);
              if (krule.fw_rc == FW_ACCEPT) {
                if ((krule.fw_flags & SF_FTP_NO_ACTIVE) == 0)
                  if ((krule.fw_flags & SF_FTP_NO_PASSIVE) == 0)
                    printf(" `ftp_all'");
                  else
                    printf(" `ftp_active'");
                else
                  if ((krule.fw_flags & SF_FTP_NO_PASSIVE) != 0)
                    printf(" `ftp_none'");
                  else
                    printf(" `ftp_passive'");       
                if ((krule.fw_flags & SF_FTP_DATA_LOG) != 0)
                  printf(" `ftp_log_data_conn'");
                if ((krule.fw_flags & SF_CHECK_SEQUENCE) != 0)
                  printf(" `check_sequence'");
              }
              printf("\n");
              tcp_or_udp = 1;
            }
	    else {
	      printf("    protocol %i (", krule.protocol);
	      if (krule.protocol <= MAX_PROTOCOL)
		printf("%s)\n", protocol[krule.protocol]);
	      else
		printf("Unknown)\n");

	      if ((krule.protocol == IPPROTO_TCP) || (krule.protocol == IPPROTO_UDP))
		tcp_or_udp = 1;
	    }
	  }
	}
        else { /* all protocols */
          printf("    all protocols");
          if (krule.fw_rc == FW_ACCEPT) {
            if ((krule.fw_flags & SF_FTP_NO_ACTIVE) == 0)
              if ((krule.fw_flags & SF_FTP_NO_PASSIVE) == 0)
                printf(" `ftp_all'");
              else
                printf(" `ftp_active'");
            else
              if ((krule.fw_flags & SF_FTP_NO_PASSIVE) != 0)
                printf(" `ftp_none'");
              else
                printf(" `ftp_passive'");
            if ((krule.fw_flags & SF_FTP_DATA_LOG) != 0)
              printf(" `ftp_log_data_conn'");
            if ((krule.fw_flags & SF_CHECK_SEQUENCE) != 0)
              printf(" `check_sequence'");
          }
          printf("\n");
        }

	if ((krule.fw_flags & SF_FW_CHECK_TTL) != 0) {
	  printf("    ttl field ");
	  if ((krule.fw_flags & SF_FW_TTL) == SF_FW_TTL_EQUAL)
	    printf("-");
	  else if ((krule.fw_flags & SF_FW_TTL) == SF_FW_TTL_LESS)
	    printf("<");
	  else if ((krule.fw_flags & SF_FW_TTL) == SF_FW_TTL_GREATER)
	    printf(">");
	  else if ((krule.fw_flags & SF_FW_TTL) == SF_FW_TTL_NOTEQUAL)
	    printf("!=");
	  else
	    printf("?");
	  printf(" %i\n", krule.ttl);
	}

	if ((krule.fw_flags & SF_FW_CHECK_OPT) != 0) {
	  printf("    options: ");
	  if ((krule.fw_flags & SF_FW_OPT_RR) != 0)
	    printf("`record route' ");
	  if ((krule.fw_flags & SF_FW_OPT_TS) != 0)
	    printf("`time stamp' ");
	  if ((krule.fw_flags & SF_FW_OPT_SEC) != 0)
	    printf("`security' ");
	  if ((krule.fw_flags & SF_FW_OPT_LSR) != 0)
	    printf("`loose source route' ");
	  if ((krule.fw_flags & SF_FW_OPT_SSR) != 0)
	    printf("`strict source route' ");
	  if ((krule.fw_flags & SF_FW_OPT_SATID) != 0)
	    printf("`sat id'");
	  printf("\n");
	}

	if (krule.fw_src_cnt) {
	  if ((krule.fw_flags & SF_FW_SRC_NEG) != 0)
	    printf("    not from ");
	  else
	    printf("    from     ");
	  print_addr(krule.fw_src_idx, krule.fw_src_cnt, tcp_or_udp,
		     "             ", krule.protocol);
	}
	if (krule.fw_dst_cnt) {
	  if ((krule.fw_flags & SF_FW_DST_NEG) != 0)
	    printf("    not to   ");
	  else
	    printf("    to       ");
	  print_addr(krule.fw_dst_idx, krule.fw_dst_cnt, tcp_or_udp,
		     "             ", krule.protocol);
	}
      }
    } while (result > 0);	/* signals end of data with 0 length data */
  }				/* daemon running */

  /* read open TCP connections */

  sysinfo(&info);
  now = time(NULL);

  printf("\n\nTCP connections at %s:\n", print_time(info.uptime, info.uptime, now));
  printf("----------------------------\n");
  printf("id/key   created  last use timeout  state/hosts client          server\n");
  printf("-------- -------- -------- -------- ----------- sequence window sequence window");

  do {
    struct sf_tcp_connection conn;
    struct servent *serv_ent;
    char source_port[80], target_port[80];
    char client[20], server[20];
    char time_established[20], time_lastused[20];

    result = longatomicread(fd, &conn, sizeof(conn));
    if (result < 0) {
      perror("Error reading the device");
      exit(1);
    }

    if (result > 0) {
      __u32 key;

      if (conn.allpsource)
	sprintf(source_port, "*");
      else {
	if ((ntohs(conn.client_port) == FTP_DATA_PORT) &&
	    ((serv_ent = getservbyport(conn.client_port, "tcp")) != NULL))
	  sprintf(source_port, "%s", serv_ent->s_name);
	else
	  sprintf(source_port, "%i", ntohs(conn.client_port));
      }

      if ((ntohs(conn.client_port) != FTP_DATA_PORT) &&
	  ((serv_ent = getservbyport(conn.server_port, "tcp")) != NULL))
	sprintf(target_port, "%s", serv_ent->s_name);
      else
	sprintf(target_port, "%i", ntohs(conn.server_port));

      strcpy(client, inet_ntoa(*((struct in_addr *) &(conn.client_addr))));
      strcpy(server, inet_ntoa(*((struct in_addr *) &(conn.server_addr))));
      strcpy(time_established, print_time(conn.established / HZ, info.uptime, now));
      strcpy(time_lastused, print_time(conn.lastused / HZ, info.uptime, now));

      key = SF_TCP_KEY(conn.client_addr + conn.server_addr +
	    (__u32) conn.client_port + (__u32) conn.server_port),

      printf("\n%08x %08x %08x %08lx %s %08x %06x %08x %06x\n    %04x %-8s %-8s %-8s %s:%s->%s:%s",
	     conn.id,
	     conn.established, conn.lastused,
	     conn.timer_set ? conn.timer.expires : 0,
	     print_state(conn.state),
	     conn.client_seq, conn.client_win << conn.client_scale,
	     conn.server_seq, conn.server_win << conn.server_scale, 
             key, time_established, time_lastused,
	     print_diff(conn.timer_set ? conn.timer.expires / HZ : 0, info.uptime),
	     client, source_port, server, target_port);

      if (idlekill && conn.lastused / HZ + idlekill < info.uptime)
        if (tcpkill(conn.id, key, conn.established) > 0)
         printf("\n    TERMINATED.");
    }
  } while (result > 0);

  close(fd);

  time(&ltime);

  if (pid && !tcponly) {	/* daemon is running */

    printf("\n\nVariables:\n");
    printf("----------\n\n");

    for (i = 0; i < sf_var_free; i++) {

      if (varnames[i] != NULL) {
	time_t timeout;

	timeout = variables[i].timeout;
	printf("%s = %li", varnames[i], (timeout > ltime) ? variables[i].value : 0);
	if ((timeout > ltime) && (timeout < INT_MAX)) {
	  strftime(buffer, sizeof(buffer), TIME_FORMAT, localtime(&timeout));
	  printf(", timeout %s\n", buffer);
	}
	else
	  printf("\n");

	if (variables[i].next != NULL) {
	  struct variable *v;

	  v = variables[i].next;
	  while (v != NULL) {
	    timeout = v->timeout;
	    if (timeout > ltime) {
	      printf("    host %-15s = %li",
	       inet_ntoa(*((struct in_addr *) &(v->address))), v->value);
	      if ((timeout > ltime) && (timeout < INT_MAX)) {
		strftime(buffer, sizeof(buffer), TIME_FORMAT, localtime(&timeout));
		printf(", timeout %s\n", buffer);
	      }
	      else
		printf("\n");
	    }

	    v = v->next;
	  }
	}

	printf("\n");
      }
    }
  }				/* daemon running */

  printf("\n");
}

static void print_addr(unsigned long index, unsigned long count,
		       int tcp_or_udp, char *s, int protocol)
{
  struct in_addr inadr;
  unsigned long k;

  for (k = index; k < (index + count); k++) {
    if (k > index)
      printf(s);
    if (sf_addr[k].mask != SF_IGNORE_ADDRESS) {
      inadr.s_addr = sf_addr[k].addr;
      printf(inet_ntoa(inadr));
      if (sf_addr[k].mask != 0xffffffff) {
	inadr.s_addr = sf_addr[k].mask;
	printf(" mask %s ", inet_ntoa(inadr));
      }
      else
	printf(" ");
    }
    if (tcp_or_udp && sf_addr[k].port != 0 && sf_addr[k].prend != USHRT_MAX) {
      printf("port %hu", sf_addr[k].port);
      if (sf_addr[k].prend != sf_addr[k].port)
	printf("..%hu", sf_addr[k].prend);
      else {
	struct servent *serv_ent;
	char proto[4];

	if (protocol == IPPROTO_UDP)
	  strcpy(proto, "udp");
	else
	  strcpy(proto, "tcp");

	if ((serv_ent = getservbyport(htons(sf_addr[k].port), proto)) != NULL)
	  printf(" (%s)", serv_ent->s_name);
      }
    }
    printf("\n");
  }
}

static char *print_state(int state)
{
  static char buffer[20];
  char *s;

  switch (state) {
    case SF_TCP_ACCEPT_SYN:
      s = "NO SYNS YET";
      break;
    case SF_TCP_CLIENT_SYN:
      s = "CLIENT_SYN ";
      break;
    case SF_TCP_SYN_ACK:
      s = "SYN_ACK    ";
      break;
    case SF_TCP_ESTABLISHED3:
      s = "ESTABLISHED";
      break;
    case SF_TCP_ESTABLISHEDFTP:
      s = "ESTABL. FTP";
      break;
    case SF_TCP_CLIENT_FIN | SF_TCP_SYN_ACK:
    case SF_TCP_CLIENT_FIN | SF_TCP_ESTABLISHED3:
    case SF_TCP_CLIENT_FIN | SF_TCP_ESTABLISHEDFTP:
      s = "CLIENT_FIN ";
      break;
    case SF_TCP_SERVER_FIN | SF_TCP_SYN_ACK:
    case SF_TCP_SERVER_FIN | SF_TCP_ESTABLISHED3:
    case SF_TCP_SERVER_FIN | SF_TCP_ESTABLISHEDFTP:
      s = "SERVER_FIN ";
      break;
    case SF_TCP_TERMINATED & ~SF_TCP_SYN_ACK:
    case SF_TCP_TERMINATED:
    case SF_TCP_TERMINATED | SF_TCP_FTP:
      s = "TERMINATED ";
      break;
    default:
      s = "??? huh ???";
      break;
  }

  strcpy(buffer, s);
  return buffer;
}

static char *print_time(time_t t, time_t up, time_t now)
{
  static char buffer[30];
  time_t event = now - up + t;

  /* jdiff == seconds since 1970 when jiffies were 0 */

  if (t == 0)
    strcpy(buffer, "-none-");
  else {
    struct tm *local;

    local = localtime(&event);
    if (abs(t - up) >= 7 * 24 * 24 * 60) {
      /* print month, day */
      strftime(buffer, sizeof(buffer), "%b %d", local);
    }
    else if (abs(t - up) >= 24 * 60 * 60) {
      /* print weekday, hour:minute */
      strftime(buffer, sizeof(buffer), "%a%H:%M", local);
    }
    else {
      /* print hour:minute:second */
      strftime(buffer, sizeof(buffer), "%H:%M:%S", local);
    }
  }

  return buffer;
}


static char *print_diff(time_t t, time_t up)
{
  static char buffer[20];

  if (t == 0)
    strcpy(buffer, "-none-");
  else {
    if (t <= up)
      strcpy(buffer, "now");
    else
      sprintf(buffer, "%lu sec", t - up);
  }

  return buffer;
}


/*
 * Kill specified TCP connection
 */

static void tcpkillidle(char *minutes)
{
  __u32 time;

  if (sscanf(minutes, "%u", &time) != 1 || !time) {
    fprintf(stderr, "%s is not a valid idle time!\n", minutes);
    exit(1);
  }

  showcfg_or_kill(1, time*60);
}


static void tcpkillarg(char *ascid, char *asckey, char *asccreated)
{
  int rc;
  __u32 id, key, created;

  if (sscanf(ascid, "%08x", &id) != 1) {
    fprintf(stderr, "%s is not a valid connection id!\n", ascid);
    exit(1);
  }
  if (sscanf(asckey, "%04x", &key) != 1) {
    fprintf(stderr, "%s is not a valid connection key!\n", asckey);
    exit(1);
  }
  if (sscanf(asccreated, "%08x", &created) != 1) {
    fprintf(stderr, "%s is not a valid connection establishment time!\n", asccreated);
    exit(1);
  }

  printf("Searching for TCP connection %08x %04x %08x...\n", id, key, created);

  rc = tcpkill(id, key, created);

  if (rc == -1)
    printf("Connection not found.\n");
  else if (rc > 0)
    printf("Connection terminated.\n");
  else
    printf("Unknown result.\n");
}


static int tcpkill(__u32 id, __u32 key, __u32 created)
{
  int pid, i, fd;
  int rc = -2;
  struct sf_control *control;
  struct sf_control_tcp *ctl;

  if ((pid = running()) != 0) {
    /* Talk to the daemon */

    if ((ctl = (struct sf_control_tcp *) malloc(sizeof(struct sf_control_tcp))) == NULL) {
      fputs("Out of memory (malloc:ctl)!\n", stderr);
      exit(1);
    }
    ctl->id = id;
    ctl->key = key;
    ctl->established = created;

    kill(pid, SIGUSR2);		/* signal the process to fetch/write data */

    if ((fd = open(IPIPE, O_WRONLY)) < 0) {
      perror("Cannot open the firewall input pipe");
      exit(1);
    }
    i = SF_CMD_DAEMON_MAGIC;
    write(fd, &i, sizeof(i));
    i = SF_CMD_DAEMON_TCPKILL;
    write(fd, &i, sizeof(i));
    write(fd, ctl, sizeof(struct sf_control_tcp));

    close(fd);
    free(ctl);

    if ((fd = open(OPIPE, O_RDONLY)) < 0) {
      perror("Cannot open the firewall output pipe");
      exit(1);
    }

    /* get the result */
    longatomicread(fd, &rc, sizeof(rc));
    close(fd);
  }
  else {
    /* Talk to kernel directly */

    if ((fd = open(sf_devname, O_WRONLY)) < 0) {
      if (errno == ENODEV)
	fputs("The module sf.o is not loaded!\n", stderr);
      else
	perror("Cannot open the firewall device");
      exit(1);
    }

    if ((control = (struct sf_control *) malloc(sizeof(struct sf_control) +
			       sizeof(struct sf_control_tcp))) == NULL) {
      fputs("Out of memory (malloc:control)!\n", stderr);
      exit(1);
    }

    control->magic = SF_CONTROL_MAGIC;
    control->command = SF_COMMAND_TCP_KILL;
    ctl = (struct sf_control_tcp *) control->data;

    ctl->id = id;
    ctl->key = key;
    ctl->established = created;

    rc = write(fd, control, sizeof(struct sf_control) + sizeof(struct sf_control_tcp));

    free(control);
    close(fd);
  }

  return rc;
}


/*
 * Kill all TCP connections
 */

static void tcpkillall(void)
{
  int pid, i, fd;

  if ((pid = running()) != 0) {
    /* Talk to the daemon */

    printf("Terminating all TCP connections...\n");

    kill(pid, SIGUSR2);		/* signal the process to fetch/write data */

    if ((fd = open(IPIPE, O_WRONLY)) < 0) {
      perror("Cannot open the firewall input pipe");
      exit(1);
    }
    i = SF_CMD_DAEMON_MAGIC;
    write(fd, &i, sizeof(i));
    i = SF_CMD_DAEMON_FLUSH_ALL;
    write(fd, &i, sizeof(i));

    close(fd);
  }
  else {
    fprintf(stderr, "The firewall daemon is not running!\n");
    exit(1);
  }
}


/*
 * Flush usage counters
 */

static void resetusage(void)
{
  int pid, i, fd;

  if ((pid = running()) != 0) {
    /* Talk to the daemon */

    printf("Resetting usage counters...\n");

    kill(pid, SIGUSR2);		/* signal the process to fetch/write data */

    if ((fd = open(IPIPE, O_WRONLY)) < 0) {
      perror("Cannot open the firewall input pipe");
      exit(1);
    }
    i = SF_CMD_DAEMON_MAGIC;
    write(fd, &i, sizeof(i));
    i = SF_CMD_DAEMON_RESETUSAGE;
    write(fd, &i, sizeof(i));

    close(fd);
  }
  else {
    fprintf(stderr, "The firewall daemon is not running!\n");
    exit(1);
  }
}


/* This function was written by Marcus Breitenstein
 * <admin@mvt.tu-cottbus.de>.
 *
 * The atomic transmission of data through pipes is limited to
 * a maximum of PIPE_BUF (==4096) bytes which is defined in the
 * linux/limits.h.
 * If SF_ADDRESS_CNT_MAX > 341 more than PIPE_BUF bytes will
 * be transfered at once during the "sfc show"-command which will
 * corrupt the datastructure.
 * The is longatomicread (int,void*,int) function is a replacement
 * for read (int,void*,int) which ensures the proper transfer of
 * larger datablocks.
 */

int longatomicread (int fd, void *pvBuff, int iBytes)
{
  if (iBytes > PIPE_BUF)
  {
    char *pvcBuff;

    int icrd;
    int ird;

    icrd = 0;
    pvcBuff = (char *)pvBuff;

    while (icrd < iBytes)
    {
      ird      = read (fd, pvcBuff, iBytes - icrd);
      if (ird<=0) return ird;
      icrd    += ird;
      pvcBuff += ird;
    }

    return icrd;
  }
  else
  {
    return read (fd, pvBuff, iBytes);
  }
}
